//
//  AKDynamicRangeCompressor.swift
//  AudioKit
//
//  Created by Aurelius Prochazka, revision history on Github.
//  Copyright © 2018 AudioKit. All rights reserved.
//

/// Dynamic range compressor from Faust
///
open class AKDynamicRangeCompressor: AKNode, AKToggleable, AKComponent, AKInput {
    public typealias AKAudioUnitType = AKDynamicRangeCompressorAudioUnit
    /// Four letter unique description of the node
    public static let ComponentDescription = AudioComponentDescription(effect: "cpsr")

    // MARK: - Properties
    private var internalAU: AKAudioUnitType?
    private var token: AUParameterObserverToken?

    fileprivate var ratioParameter: AUParameter?
    fileprivate var thresholdParameter: AUParameter?
    fileprivate var attackTimeParameter: AUParameter?
    fileprivate var releaseTimeParameter: AUParameter?

    /// Lower and upper bounds for Ratio
    public static let ratioRange = 0.01 ... 100.0

    /// Lower and upper bounds for Threshold
    public static let thresholdRange = -100.0 ... 0.0

    /// Lower and upper bounds for Attack Time
    public static let attackTimeRange = 0.0 ... 1.0

    /// Lower and upper bounds for Release Time
    public static let releaseTimeRange = 0.0 ... 1.0

    /// Initial value for Ratio
    public static let defaultRatio = 1.0

    /// Initial value for Threshold
    public static let defaultThreshold = 0.0

    /// Initial value for Attack Time
    public static let defaultAttackTime = 0.1

    /// Initial value for Release Time
    public static let defaultReleaseTime = 0.1

    /// Ramp Time represents the speed at which parameters are allowed to change
    @objc open dynamic var rampTime: Double = AKSettings.rampTime {
        willSet {
            internalAU?.rampTime = newValue
        }
    }

    /// Ratio to compress with, a value > 1 will compress
    @objc open dynamic var ratio: Double = defaultRatio {
        willSet {
            if ratio == newValue {
                return
            }
            if internalAU?.isSetUp ?? false {
                if let existingToken = token {
                    ratioParameter?.setValue(Float(newValue), originator: existingToken)
                    return
                }
            }
            internalAU?.setParameterImmediately(.ratio, value: newValue)
        }
    }

    /// Threshold (in dB) 0 = max
    @objc open dynamic var threshold: Double = defaultThreshold {
        willSet {
            if threshold == newValue {
                return
            }
            if internalAU?.isSetUp ?? false {
                if let existingToken = token {
                    thresholdParameter?.setValue(Float(newValue), originator: existingToken)
                    return
                }
            }
            internalAU?.setParameterImmediately(.threshold, value: newValue)
        }
    }

    /// Attack time
    @objc open dynamic var attackTime: Double = defaultAttackTime {
        willSet {
            if attackTime == newValue {
                return
            }
            if internalAU?.isSetUp ?? false {
                if let existingToken = token {
                    attackTimeParameter?.setValue(Float(newValue), originator: existingToken)
                    return
                }
            }
            internalAU?.setParameterImmediately(.attackTime, value: newValue)
        }
    }

    /// Release time
    @objc open dynamic var releaseTime: Double = defaultReleaseTime {
        willSet {
            if releaseTime == newValue {
                return
            }
            if internalAU?.isSetUp ?? false {
                if let existingToken = token {
                    releaseTimeParameter?.setValue(Float(newValue), originator: existingToken)
                    return
                }
            }
            internalAU?.setParameterImmediately(.releaseTime, value: newValue)
        }
    }

    /// Tells whether the node is processing (ie. started, playing, or active)
    @objc open dynamic var isStarted: Bool {
        return internalAU?.isPlaying ?? false
    }

    // MARK: - Initialization

    /// Initialize this compressor node
    ///
    /// - Parameters:
    ///   - input: Input node to process
    ///   - ratio: Ratio to compress with, a value > 1 will compress
    ///   - threshold: Threshold (in dB) 0 = max
    ///   - attackTime: Attack time
    ///   - releaseTime: Release time
    ///
    @objc public init(
        _ input: AKNode? = nil,
        ratio: Double = defaultRatio,
        threshold: Double = defaultThreshold,
        attackTime: Double = defaultAttackTime,
        releaseTime: Double = defaultReleaseTime
        ) {

        self.ratio = ratio
        self.threshold = threshold
        self.attackTime = attackTime
        self.releaseTime = releaseTime

        _Self.register()

        super.init()
        AVAudioUnit._instantiate(with: _Self.ComponentDescription) { [weak self] avAudioUnit in
            guard let strongSelf = self else {
                AKLog("Error: self is nil")
                return
            }
            strongSelf.avAudioNode = avAudioUnit
            strongSelf.internalAU = avAudioUnit.auAudioUnit as? AKAudioUnitType
            input?.connect(to: strongSelf)
        }

        guard let tree = internalAU?.parameterTree else {
            AKLog("Parameter Tree Failed")
            return
        }

        ratioParameter = tree["ratio"]
        thresholdParameter = tree["threshold"]
        attackTimeParameter = tree["attackTime"]
        releaseTimeParameter = tree["releaseTime"]

        token = tree.token(byAddingParameterObserver: { [weak self] _, _ in

            guard let _ = self else {
                AKLog("Unable to create strong reference to self")
                return
            } // Replace _ with strongSelf if needed
            DispatchQueue.main.async {
                // This node does not change its own values so we won't add any
                // value observing, but if you need to, this is where that goes.
            }
        })

        internalAU?.setParameterImmediately(.ratio, value: ratio)
        internalAU?.setParameterImmediately(.threshold, value: threshold)
        internalAU?.setParameterImmediately(.attackTime, value: attackTime)
        internalAU?.setParameterImmediately(.releaseTime, value: releaseTime)
    }

    // MARK: - Control

    /// Function to start, play, or activate the node, all do the same thing
    @objc open func start() {
        internalAU?.start()
    }

    /// Function to stop or bypass the node, both are equivalent
    @objc open func stop() {
        internalAU?.stop()
    }
}
